import { Connection, Signer, Transaction, TransactionInstruction, PublicKey, SimulatedTransactionResponse, VersionedTransaction, AddressLookupTableAccount } from "@solana/web3.js";
import { GaslessDapp } from "../dapp";
import { GaslessTypes, SignedPuzzle } from "./types";
import { Program } from "@project-serum/anchor";
import { Wallet } from "../helpers";
import { GaslessProgram } from "../dapp/artifacts/gasless_program";
export type NeededInitAccount = {
    newAccount: PublicKey;
    feePayer: string;
    lamports: number;
};
export type CompressedIx = {
    instructions: TransactionInstruction[];
    cleanupInstructions: TransactionInstruction[];
    signers: Signer[];
};
export declare class GaslessTransaction {
    readonly connection: Connection;
    readonly wallet: Wallet;
    readonly dapp: GaslessDapp;
    transaction: Transaction;
    signers: Signer[];
    gaslessType: GaslessTypes;
    gaslessProgram: Program<GaslessProgram>;
    lookupTables: AddressLookupTableAccount[] | undefined;
    v0: VersionedTransaction | null;
    constructor(connection: Connection, wallet: Wallet, dapp: GaslessDapp, gaslessType?: GaslessTypes);
    static fromTransactionBuilder(connection: Connection, wallet: Wallet, compressIx: CompressedIx, dappUtil: GaslessDapp): GaslessTransaction;
    addSigners(signers: Signer[]): GaslessTransaction;
    addInstructions(ixs: TransactionInstruction[]): GaslessTransaction;
    setGaslessType(gaslessType: GaslessTypes): GaslessTransaction;
    getPuzzleAndEstimateTime(reCaptchaToken?: string): Promise<{
        puzzle: SignedPuzzle;
        estHandlingTime: number;
    }>;
    solveAndSubmit(puzzle: SignedPuzzle): Promise<string>;
    simulateGaslessTx(includeAccounts?: PublicKey[]): Promise<SimulatedTransactionResponse>;
    build(sign?: boolean): Promise<Transaction | VersionedTransaction>;
    buildAndExecute(reCaptchaToken?: string): Promise<string>;
    solveAndSubmitPuzzle(feePayer: PublicKey, puzzle: SignedPuzzle): Promise<string>;
    private _solveAndSubmitPuzzle;
    private _fundInitAccounts;
    private _fundInitAccountsV0;
    asyncBuildAndExecute(cb: (error: any, txid: string) => void): void;
    toV0Tx(transaction: Transaction, sign: boolean | undefined, lookupTables: AddressLookupTableAccount[] | undefined): Promise<VersionedTransaction>;
    static fromV0Tx(connection: Connection, wallet: Wallet, dappUtil: GaslessDapp, v0: VersionedTransaction, signers: Signer[]): Promise<GaslessTransaction>;
}
