"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GaslessTransaction = void 0;
const web3_js_1 = require("@solana/web3.js");
const types_1 = require("./types");
const gasless_1 = require("../gasless");
const pow_1 = require("../pow");
const anchor_1 = require("@project-serum/anchor");
const helpers_1 = require("../helpers");
const gasless_program_json_1 = __importDefault(require("../dapp/artifacts/gasless_program.json"));
const spl_token_1 = require("@solana/spl-token");
// TODO: get from gasless service
const gaslessProgramId = new web3_js_1.PublicKey("GasP6kcNpTdXA1M7ENyh5kCEvtHPgy71Habxe62gqHqH");
const CLOSE_ACCOUNT_IX = 9;
const TOKEN_ACCOUNT_RENT = 2039280;
class GaslessTransaction {
    constructor(connection, wallet, dapp, gaslessType = types_1.GaslessTypes.POW) {
        this.connection = connection;
        this.wallet = wallet;
        this.dapp = dapp;
        this.lookupTables = undefined;
        this.v0 = null;
        this.transaction = new web3_js_1.Transaction();
        this.signers = [];
        this.gaslessType = gaslessType;
        const provider = new anchor_1.AnchorProvider(connection, wallet, { commitment: "confirmed" });
        const gaslessProgram = new anchor_1.Program(gasless_program_json_1.default, gaslessProgramId, provider);
        this.gaslessProgram = gaslessProgram;
    }
    static fromTransactionBuilder(connection, wallet, compressIx, dappUtil) {
        const { instructions, cleanupInstructions, signers } = compressIx;
        const _this = new GaslessTransaction(connection, wallet, dappUtil);
        _this.addSigners(signers);
        _this.addInstructions(instructions);
        _this.addInstructions(cleanupInstructions);
        return _this;
    }
    addSigners(signers) {
        this.signers = signers;
        return this;
    }
    addInstructions(ixs) {
        for (let i = 0; i < ixs.length; i++) {
            const ix = ixs[i];
            this.transaction.add(ix);
        }
        return this;
    }
    setGaslessType(gaslessType) {
        this.gaslessType = gaslessType;
        return this;
    }
    getPuzzleAndEstimateTime(reCaptchaToken) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.gaslessType !== types_1.GaslessTypes.POW) {
                throw Error(`${this.gaslessType} not supported`);
            }
            const puzzle = yield (0, gasless_1.getPuzzle)(this.connection, this.wallet.publicKey, reCaptchaToken);
            return { puzzle, estHandlingTime: pow_1.POWPuzzle.estHandlingTime(puzzle) };
        });
    }
    solveAndSubmit(puzzle) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.gaslessType !== types_1.GaslessTypes.POW) {
                throw Error(`${this.gaslessType} not supported`);
            }
            const { feePayer } = yield (0, gasless_1.getGaslessInfo)(this.connection);
            const txid = yield this._solveAndSubmitPuzzle(feePayer, puzzle);
            return txid;
        });
    }
    simulateGaslessTx(includeAccounts = []) {
        return __awaiter(this, void 0, void 0, function* () {
            const transaction = yield this.build();
            if (transaction instanceof web3_js_1.VersionedTransaction) {
                const simulated = yield this.connection.simulateTransaction(transaction, {
                    accounts: {
                        encoding: "base64",
                        addresses: includeAccounts.map((v) => v.toBase58()),
                    },
                });
                return simulated.value;
            }
            const simulated = yield (0, helpers_1.simulateRawTransaction)(this.connection, transaction, includeAccounts);
            return simulated;
        });
    }
    build(sign = false) {
        return __awaiter(this, void 0, void 0, function* () {
            // Automatically detect the dapp type
            // Be disabled automatically detect the dapp type
            // if (this.dapp.hasDappInstruction(this.transaction)) {
            //   this.gaslessType = GaslessTypes.Dapp;
            // }
            const { feePayer } = yield (0, gasless_1.getGaslessInfo)(this.connection);
            this.transaction.feePayer = feePayer;
            this.transaction.recentBlockhash = (yield this.connection.getRecentBlockhash()).blockhash;
            let fundInitAccounts;
            // Fund init accounts
            if (this.v0) {
                fundInitAccounts = yield this._fundInitAccountsV0(this.transaction, feePayer);
            }
            else {
                fundInitAccounts = yield this._fundInitAccounts(this.transaction, feePayer);
            }
            // fundInitAccounts = await this._fundInitAccounts(this.transaction, feePayer);
            const { transaction, neededInits } = fundInitAccounts;
            if (neededInits.length > 0) {
                // START: add verifyRentAccounts
                const fundAmounts = neededInits.map((v) => new anchor_1.BN(v.lamports));
                const accounts = neededInits.map((v) => {
                    return { pubkey: v.newAccount, isWritable: false, isSigner: false };
                });
                const coder = new anchor_1.BorshInstructionCoder(gasless_program_json_1.default);
                const data = coder.encode("verifyRentAccounts", { fees: fundAmounts, powSignature: "" });
                const ix = new web3_js_1.TransactionInstruction({
                    keys: [
                        {
                            pubkey: feePayer,
                            isWritable: true,
                            isSigner: true,
                        },
                        {
                            pubkey: web3_js_1.SYSVAR_RENT_PUBKEY,
                            isWritable: false,
                            isSigner: false,
                        },
                        ...accounts,
                    ],
                    programId: gaslessProgramId,
                    data,
                });
                transaction.add(ix);
                this.transaction = transaction;
                // For TEST: Simulate here
                // const finalSimulated = await simulateRawTransaction(this.connection, this.transaction, false);
                // console.log("finalSimulated", finalSimulated);
                // END: add verifyRentAccounts
            }
            if (this.v0) {
                const v0Tx = yield this.toV0Tx(this.transaction, sign, this.lookupTables);
                return v0Tx;
            }
            else {
                if (sign) {
                    for (let i = 0; i < this.signers.length; i++) {
                        const s = this.signers[i];
                        this.transaction.sign(s);
                    }
                    this.transaction = yield this.wallet.signTransaction(this.transaction);
                }
                return this.transaction;
            }
        });
    }
    buildAndExecute(reCaptchaToken) {
        return __awaiter(this, void 0, void 0, function* () {
            // Be disabled automatically detect the dapp type
            // if (this.dapp.hasDappInstruction(this.transaction)) {
            //   this.gaslessType = GaslessTypes.Dapp;
            // }
            const { feePayer } = yield (0, gasless_1.getGaslessInfo)(this.connection);
            if (this.gaslessType === types_1.GaslessTypes.POW) {
                const puzzle = yield (0, gasless_1.getPuzzle)(this.connection, this.wallet.publicKey, reCaptchaToken);
                const txid = yield this._solveAndSubmitPuzzle(feePayer, puzzle);
                return txid;
            }
            else {
                throw Error(`${this.gaslessType} not supported`);
            }
        });
    }
    solveAndSubmitPuzzle(feePayer, puzzle) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._solveAndSubmitPuzzle(feePayer, puzzle);
        });
    }
    _solveAndSubmitPuzzle(feePayer, puzzle) {
        return __awaiter(this, void 0, void 0, function* () {
            const solution = yield pow_1.POWPuzzle.solveAsync(pow_1.Question.fromObject(puzzle.question));
            const rawSolution = Object.assign({ address: this.wallet.publicKey.toBase58(), solution: solution.toString(16) }, puzzle);
            // check if we can submit solution at this point (now >= puzzle.allowedSubmissionAt)
            const now = Math.floor(Date.now() / 1000);
            if (puzzle.allowedSubmissionAt && now < puzzle.allowedSubmissionAt) {
                yield (0, helpers_1.sleep)((puzzle.allowedSubmissionAt - now) * 1000);
            }
            // build transaction
            const transaction = yield this.build(true);
            const txid = yield (0, gasless_1.postSolution)(this.connection, rawSolution, transaction);
            return txid;
        });
    }
    _fundInitAccounts(transaction, gaslessFeePayer) {
        return __awaiter(this, void 0, void 0, function* () {
            const initAccounts = [];
            for (let ix of transaction.instructions) {
                const feePayer = ix.keys.find((k) => k.isSigner);
                const keys = ix.keys.map((value) => value.pubkey);
                const infos = yield this.connection.getMultipleAccountsInfo(keys);
                // In transaction, the first signer must be feePayer of DApps and ignore the gasless's feePayer.
                if (feePayer && feePayer.pubkey.toBase58() !== gaslessFeePayer.toBase58()) {
                    for (let i = 0; i < keys.length; i++) {
                        const accountMeta = ix.keys[i];
                        const info = infos[i];
                        const isExisted = initAccounts.find((v) => v.newAccount.toBase58() === accountMeta.pubkey.toBase58());
                        if (!info && !isExisted) {
                            initAccounts.push({
                                newAccount: accountMeta.pubkey,
                                feePayer: feePayer.pubkey.toBase58(),
                                lamports: 0,
                            });
                        }
                    }
                }
            }
            if (initAccounts.length > 0) {
                const ixs = transaction.instructions;
                let feePayer = null;
                const grouped = groupByToMap(initAccounts, (v) => v.feePayer);
                if (grouped.size > 1) {
                    throw "Only support a single fee payer";
                }
                grouped.forEach((v, k) => {
                    feePayer = new web3_js_1.PublicKey(k);
                    ixs.unshift(web3_js_1.SystemProgram.transfer({
                        fromPubkey: gaslessFeePayer,
                        toPubkey: feePayer,
                        lamports: 100000000, // 0.1 RENECs
                    }));
                });
                if (feePayer) {
                    transaction.instructions = ixs;
                    const simulated = yield (0, helpers_1.simulateRawTransaction)(this.connection, transaction, initAccounts.map((v) => v.newAccount));
                    if (simulated.err) {
                        const message = { err: simulated.err, logs: simulated.logs };
                        throw "simulated error:" + JSON.stringify(message);
                    }
                    if (simulated.accounts && simulated.accounts.length > 0) {
                        let totalLamports = 0;
                        for (let i = 0; i < initAccounts.length; i++) {
                            const account = simulated.accounts[i];
                            if (account && account.owner !== web3_js_1.SystemProgram.programId.toBase58()) {
                                const data = Buffer.from(account.data[0], "base64");
                                const lamports = yield this.connection.getMinimumBalanceForRentExemption(data.byteLength);
                                initAccounts[i].lamports = lamports;
                                totalLamports += lamports;
                            }
                        }
                        const ixs = transaction.instructions;
                        ixs.shift();
                        const closeAccountIx = ixs.find((ix) => {
                            return ix.programId.equals(spl_token_1.TOKEN_PROGRAM_ID) && ix.data.readInt8() === CLOSE_ACCOUNT_IX;
                        });
                        // needs to fund lamports for closed token account.
                        if (closeAccountIx) {
                            const closedAccount = closeAccountIx.keys[0].pubkey.toBase58();
                            for (let i = 0; i < initAccounts.length; i++) {
                                if (initAccounts[i].newAccount.toBase58() === closedAccount) {
                                    initAccounts[i].lamports = TOKEN_ACCOUNT_RENT;
                                    totalLamports += TOKEN_ACCOUNT_RENT;
                                }
                            }
                        }
                        if (totalLamports > 0) {
                            ixs.unshift(web3_js_1.SystemProgram.transfer({
                                fromPubkey: gaslessFeePayer,
                                toPubkey: feePayer,
                                lamports: totalLamports,
                            }));
                        }
                        transaction.instructions = ixs;
                    }
                }
            }
            return { transaction, neededInits: initAccounts.filter((v) => v.lamports > 0) };
        });
    }
    _fundInitAccountsV0(transaction, gaslessFeePayer) {
        return __awaiter(this, void 0, void 0, function* () {
            const initAccounts = [];
            for (let ix of transaction.instructions) {
                const feePayer = ix.keys.find((k) => k.isSigner);
                const keys = ix.keys.map((value) => value.pubkey);
                const infos = yield this.connection.getMultipleAccountsInfo(keys);
                // In transaction, the first signer must be feePayer of DApps and ignore the gasless's feePayer.
                if (feePayer && feePayer.pubkey.toBase58() !== gaslessFeePayer.toBase58()) {
                    for (let i = 0; i < keys.length; i++) {
                        const accountMeta = ix.keys[i];
                        const info = infos[i];
                        const isExisted = initAccounts.find((v) => v.newAccount.toBase58() === accountMeta.pubkey.toBase58());
                        if (!info && !isExisted) {
                            initAccounts.push({
                                newAccount: accountMeta.pubkey,
                                feePayer: feePayer.pubkey.toBase58(),
                                lamports: 0,
                            });
                        }
                    }
                }
            }
            if (initAccounts.length > 0) {
                const ixs = transaction.instructions;
                let feePayer = null;
                const grouped = groupByToMap(initAccounts, (v) => v.feePayer);
                if (grouped.size > 1) {
                    throw "Only support a single fee payer";
                }
                grouped.forEach((v, k) => {
                    feePayer = new web3_js_1.PublicKey(k);
                    ixs.unshift(web3_js_1.SystemProgram.transfer({
                        fromPubkey: gaslessFeePayer,
                        toPubkey: feePayer,
                        lamports: 100000000, // 0.1 RENECs
                    }));
                });
                if (feePayer) {
                    transaction.instructions = ixs;
                    if (initAccounts.length > 0) {
                        let v0 = yield this.toV0Tx(transaction, false, this.lookupTables);
                        const simulated = yield this.connection.simulateTransaction(v0, {
                            sigVerify: false,
                            accounts: {
                                encoding: "base64",
                                addresses: initAccounts.map((v) => v.newAccount.toBase58()),
                            },
                        });
                        if (simulated.value.err) {
                            throw simulated.value.err;
                        }
                        if (simulated.value.accounts && simulated.value.accounts.length > 0) {
                            let totalLamports = 0;
                            for (let i = 0; i < initAccounts.length; i++) {
                                const account = simulated.value.accounts[i];
                                if (account && account.owner !== web3_js_1.SystemProgram.programId.toBase58()) {
                                    const data = Buffer.from(account.data[0], "base64");
                                    const lamports = yield this.connection.getMinimumBalanceForRentExemption(data.byteLength);
                                    initAccounts[i].lamports = lamports;
                                    totalLamports += lamports;
                                }
                            }
                            const ixs = transaction.instructions;
                            ixs.shift();
                            if (totalLamports > 0) {
                                ixs.unshift(web3_js_1.SystemProgram.transfer({
                                    fromPubkey: gaslessFeePayer,
                                    toPubkey: feePayer,
                                    lamports: totalLamports, // 1 RENECs
                                }));
                            }
                            transaction.instructions = ixs;
                            v0 = yield this.toV0Tx(transaction, false, this.lookupTables);
                            const finalSimulated = yield this.connection.simulateTransaction(v0, {
                                sigVerify: false,
                                accounts: {
                                    encoding: "base64",
                                    addresses: initAccounts.map((v) => v.newAccount.toBase58()),
                                },
                            });
                            if (finalSimulated.value.err) {
                                throw finalSimulated.value.err;
                            }
                        }
                    }
                }
            }
            return { transaction, neededInits: initAccounts.filter((v) => v.lamports > 0) };
        });
    }
    asyncBuildAndExecute(cb) {
        setTimeout(() => __awaiter(this, void 0, void 0, function* () {
            try {
                const txid = yield this.buildAndExecute();
                cb(null, txid);
            }
            catch (e) {
                cb(e, "");
            }
        }), 0);
    }
    toV0Tx(transaction, sign = false, lookupTables) {
        return __awaiter(this, void 0, void 0, function* () {
            const feePayer = transaction.feePayer
                ? transaction.feePayer
                : (yield (0, gasless_1.getGaslessInfo)(this.connection)).feePayer;
            const latestBlockhash = transaction.recentBlockhash
                ? transaction.recentBlockhash
                : (yield this.connection.getRecentBlockhash()).blockhash;
            const messageV0 = new web3_js_1.TransactionMessage({
                payerKey: feePayer,
                recentBlockhash: latestBlockhash,
                instructions: transaction.instructions,
            }).compileToV0Message(lookupTables);
            let v0Transaction = new web3_js_1.VersionedTransaction(messageV0);
            if (sign) {
                for (let i = 0; i < this.signers.length; i++) {
                    const s = this.signers[i];
                    v0Transaction.sign([s]);
                }
                v0Transaction = yield this.wallet.signTransaction(v0Transaction);
            }
            return v0Transaction;
        });
    }
    static fromV0Tx(connection, wallet, dappUtil, v0, signers) {
        return __awaiter(this, void 0, void 0, function* () {
            const { addressTableLookups, compiledInstructions, recentBlockhash } = v0.message;
            const lookupTables = [];
            for (let i = 0; i < addressTableLookups.length; i++) {
                const table = addressTableLookups[i];
                const addrLookupTable = yield connection.getAddressLookupTable(table.accountKey);
                if (!addrLookupTable.value)
                    throw Error(`Could not find addressLookupTable ${table.accountKey.toBase58()}`);
                lookupTables.push(addrLookupTable.value);
            }
            const accounts = v0.message.getAccountKeys({ addressLookupTableAccounts: lookupTables });
            const ixs = compiledInstructions.map((ix) => {
                const programId = accounts.get(ix.programIdIndex);
                if (!programId)
                    throw Error(`Not found programId index ${ix.programIdIndex}`);
                const keys = ix.accountKeyIndexes.map((index) => {
                    const key = accounts.get(index);
                    if (!key)
                        throw Error(`Not found key index ${ix.programIdIndex}`);
                    return {
                        pubkey: key,
                        isSigner: v0.message.isAccountSigner(index),
                        isWritable: v0.message.isAccountWritable(index),
                    };
                });
                const data = Buffer.from(ix.data);
                const txInstruction = {
                    programId,
                    keys,
                    data,
                };
                return txInstruction;
            });
            const _this = new GaslessTransaction(connection, wallet, dappUtil);
            _this.addSigners(signers);
            _this.addInstructions(ixs);
            _this.v0 = v0;
            _this.lookupTables = lookupTables;
            return _this;
        });
    }
}
exports.GaslessTransaction = GaslessTransaction;
const groupByToMap = (array, predicate) => array.reduce((map, value, index, array) => {
    var _a, _b;
    const key = predicate(value, index, array);
    (_b = (_a = map.get(key)) === null || _a === void 0 ? void 0 : _a.push(value)) !== null && _b !== void 0 ? _b : map.set(key, [value]);
    return map;
}, new Map());
