"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.simulateRawTransaction = void 0;
const web3_js_1 = require("@solana/web3.js");
// Simulate a signed, serialized transaction before broadcasting
function simulateRawTransaction(connection, transaction, includeAccounts) {
    return __awaiter(this, void 0, void 0, function* () {
        /*
             Simulating a transaction directly can cause the `signatures` property to change.
             Possibly related:
             https://github.com/solana-labs/solana/issues/21722
             https://github.com/solana-labs/solana/pull/21724
             https://github.com/solana-labs/solana/issues/20743
             https://github.com/solana-labs/solana/issues/22021
      
             Clone it from the bytes instead, and make sure it's likely to succeed before paying for it.
      
             Within simulateTransaction there is a "transaction instanceof Transaction" check. Since connection is passed
             from outside the library, it uses parent application's version of web3.js. "instanceof" won't recognize a match.
             Instead, let's explicitly call for simulateTransaction within the dependency of the library.
           */
        const simulateTransaction = web3_js_1.Connection.prototype.simulateTransaction;
        const simulated = yield simulateTransaction.call(connection, transaction, undefined, includeAccounts);
        return simulated.value;
    });
}
exports.simulateRawTransaction = simulateRawTransaction;
